from packaging.requirements import Requirement
from packaging.specifiers import SpecifierSet

from ..errors import InvalidRequirementError
from .hab_base import HabBase


class Distro(HabBase):
    """Container of DistroVersion objects. One per distro exists in a distro forest"""

    def latest_version(self, specifier):
        """Returns the newest version available matching the specifier"""
        versions = self.matching_versions(specifier)
        try:
            version = max(versions)
        except ValueError:
            raise InvalidRequirementError(
                f'Unable to find a valid version for "{specifier}" in versions '
                f'[{", ".join([str(v) for v in self.versions.keys()])}]'
            ) from None
        return self.versions[version]

    def matching_versions(self, specification):
        """Returns a list of versions available matching the version specification.
        See `packaging.requirements` for details on valid requirements, but it
        should be the same as pip requirements.
        """
        if isinstance(specification, Requirement):
            specifier = specification.specifier
        elif isinstance(specification, SpecifierSet):
            specifier = specification
        else:
            specifier = Requirement(specification).specifier
        return specifier.filter(
            self.versions.keys(), prereleases=self.resolver.prereleases
        )

    @property
    def versions(self):
        """A dict of available distro versions"""
        return {c.version: c for c in self.children}



import os
from contextlib import contextmanager
from pathlib import Path, PurePath

import pytest
from packaging.requirements import Requirement

from hab import Resolver, Site


@pytest.fixture
def config_root():
    return Path(__file__).parent


@pytest.fixture
def resolver(config_root):
    """Return a standard testing resolver"""
    site = Site([config_root / "site_main.json"])
    return Resolver(site=site)


class Helpers(object):
    """A collection of reusable functions that tests can use."""

    @staticmethod
    def assert_requirements_equal(req, check):
        """Assert that a requirement dict matches a list of requirements.

        Args:
            req (dict): A Requirement dictionary matching the output of
                ``hab.solvers.Solvers.simplify_requirements``.
            check (list): A list of requirement strings. This takes a list
                so writing tests requires less boilerplate.

        Raises:
            AssertionError: If the provided req and check don't exactly match.
        """
        try:
            assert len(req) == len(check)
        except AssertionError:
            # Provide additional information to help debug a failing test. The simple
            # len assert doesn't make it easy to debug a failing test
            print(" Requirement dict ".center(50, "-"))
            print(req)
            print(" Check ".center(50, "-"))
            print(check)
            raise
        for chk in check:
            chk = Requirement(chk)
            assert Helpers.cmp_requirement(req[chk.name], chk)

    @staticmethod
    def check_path_list(paths, checks):
        """Casts the objects in both lists to PurePath objects so they can be
        reliably asserted and differences easily viewed in the pytest output.
        """
        paths = [PurePath(p) for p in paths]
        checks = [PurePath(p) for p in checks]
        assert paths == checks

    @staticmethod
    def cmp_requirement(a, b):
        """Convenience method to check if two Requirement objects are the same.

        Args:
            a (Requirement): The first Requirement to compare
            b (Requirement): The second Requirement to compare

        Returns:
            bool: If a and b represent the same requirement
        """
        return type(a) == type(b) and str(a) == str(b)

    @staticmethod
    @contextmanager
    def reset_environ():
        """Resets the environment variables once the with context exits."""
        old_environ = dict(os.environ)
        try:
            yield
        finally:
            # Restore the original environment variables
            os.environ.clear()
            os.environ.update(old_environ)


@pytest.fixture
def helpers():
    """Expose the Helpers class as a fixture for ease of use in tests."""
    return Helpers



from . import HabBase
from .meta import hab_property


class Placeholder(HabBase):
    """Provides an parent node for a child if one hasn't been created yet.
    This node will be replaced in the tree if a node is loaded for this position.
    """

    @hab_property(verbosity=2)
    def inherits(self):
        """Placeholders don't contain their own data, so they always inherit."""
        return True


# This is the first place where both HabBase and its subclass Placeholder
# are defined, so this is where we have to set _placeholder.
HabBase._placeholder = Placeholder



from packaging.version import InvalidVersion, Version

from .. import NotSet
from ..errors import InvalidVersionError, _IgnoredVersionError
from .distro import Distro
from .hab_base import HabBase
from .meta import hab_property


class DistroVersion(HabBase):
    """A specific version of the loaded `Distro`'s. Including its requirements,
    aliases and environment variables."""

    _context_method = "name"
    _placeholder = Distro

    def __init__(self, *args, **kwargs):
        self._alias_mods = NotSet
        super().__init__(*args, **kwargs)

    @hab_property()
    def aliases(self):
        """List of the names and commands that need created to launch desired
        applications."""
        return self.frozen_data.get("aliases", NotSet)

    @aliases.setter
    def aliases(self, aliases):
        self.frozen_data["aliases"] = aliases

    # Note: 'alias_mods' needs to be processed before 'environment'
    @hab_property(process_order=50)
    def alias_mods(self):
        """Dict of modifications that need to be made on aliases.
        These are used to modify the original configuration of an alias by another
        distro or config. This allows a plugin to add an environment variable to
        a specific alias even though the alias is defined by another distro/config.
        """
        return self._alias_mods

    def load(self, filename):
        # Fill in the DistroVersion specific settings before calling super
        data = self._load(filename)

        self.aliases = data.get("aliases", NotSet)
        # Store any alias_mods, they will be processed later when flattening
        self._alias_mods = data.get("alias_mods", NotSet)

        # The version can be stored in several ways to make deployment and dev easier
        version_txt = self.dirname / ".hab_version.txt"
        if "version" in data:
            self.version = data["version"]
        elif version_txt.exists():
            self.version = version_txt.open().read().strip()
        else:
            # If version is not defined in json data extract it from the parent
            # directory name. This allows for simpler distribution without needing
            # to modify version controlled files.
            try:
                self.version = self.dirname.name
            except InvalidVersion:
                """The parent directory was not a valid version, attempt to get a
                version using setuptools_scm.
                """
                try:
                    from setuptools_scm import get_version
                except ImportError as error:
                    raise InvalidVersionError(self.filename, error=error) from None

                def check_ignored_version():
                    if self.dirname.name in self.resolver.ignored:
                        # This object is not added to the forest until super is called
                        raise _IgnoredVersionError(
                            'Skipping "{}" its dirname is in the ignored list.'.format(
                                filename
                            )
                        ) from None

                try:
                    self.version = get_version(
                        root=self.dirname, version_scheme="release-branch-semver"
                    )
                except LookupError:
                    check_ignored_version()
                    raise InvalidVersionError(self.filename) from None
                except Exception as error:
                    check_ignored_version()
                    # To make debugging easier include the original exception
                    raise InvalidVersionError(self.filename, error=error) from None

        # The name should be the version == specifier.
        self.distro_name = data.get("name")
        self.name = "{}=={}".format(self.distro_name, self.version)

        data = super().load(filename, data=data)

        return data

    @hab_property()
    def version(self):
        return super().version

    @version.setter
    def version(self, version):
        # NOTE: super doesn't work for a @property.setter
        if version and not isinstance(version, Version):
            version = Version(version)
        self.frozen_data["version"] = version



from collections import OrderedDict

import pytest
from packaging.requirements import Requirement

from hab.errors import InvalidRequirementError, MaxRedirectError
from hab.solvers import Solver


@pytest.mark.parametrize(
    "value,check",
    (
        # List and tuple inputs
        (["the_dcc"], ["the_dcc"]),
        (("the_dcc", "a_dcc"), ["the_dcc", "a_dcc"]),
        (["the_dcc", "a_dcc"], ["the_dcc", "a_dcc"]),
        ([Requirement("the_dcc"), "a_dcc"], ["the_dcc", "a_dcc"]),
        # simple dict inputs
        ({"the_dcc": None, "b_dcc": None}, ["the_dcc", "b_dcc"]),
        ({"the_dcc": "c_dcc"}, ["the_dcc", "c_dcc"]),
        ({Requirement("a_dcc"): None}, ["a_dcc"]),
        ({"a_dcc": Requirement("a_dcc")}, ["a_dcc"]),
        ({Requirement("the_dcc"): Requirement("a_dcc")}, ["the_dcc", "a_dcc"]),
        # Merging requirements
        ({"the_dcc": "the_dcc>=1.0", "b_dcc": None}, ["the_dcc>=1.0", "b_dcc"]),
        ({"the_dcc>=0.9": "the_dcc>=1.0"}, ["the_dcc>=0.9,>=1.0"]),
        # Dict containing lists
        (
            {"the_dcc": ["a_dcc", "b_dcc"], "an_dcc": ["a_dcc", "c_dcc"]},
            ["the_dcc", "an_dcc", "a_dcc", "b_dcc", "c_dcc"],
        ),
    ),
)
def test_simplify_requirements(helpers, value, check):
    ret = Solver.simplify_requirements(value)
    helpers.assert_requirements_equal(ret, check)


@pytest.mark.parametrize(
    "requirements,match",
    (
        (
            {"no_existant_distro": Requirement("no_existant_distro")},
            "Unable to find a distro for requirement: no_existant_distro",
        ),
        # Testing marker output. Using Invalid so this test works on all platforms
        (
            {"no_exist": Requirement("no_exist;platform_system!='Invalid'")},
            'Unable to find a distro for requirement: no_exist; platform_system != "Invalid"',
        ),
        (
            {"the_dcc": Requirement("the_dcc==0.0.0")},
            r'Unable to find a valid version for "the_dcc==0.0.0" in versions \[.+\]',
        ),
        (
            # This requirement is not possible because the_dcc_plugin_b requires the_dcc<1.2
            {
                "the_dcc": Requirement("the_dcc>1.1"),
                "the_dcc_plugin_b": Requirement("the_dcc_plugin_b<1.0"),
            },
            r'Unable to find a valid version for "the_dcc<1.2,>1.1" in versions \[.+\]',
        ),
    ),
)
def test_invalid_requirement_errors(resolver, requirements, match):
    """Test that the correct error is raised if an invalid or missing requirement
    is specified."""
    with pytest.raises(InvalidRequirementError, match=match):
        resolver.resolve_requirements(requirements)


def test_solver_errors(resolver):
    """Test that the correct errors are raised"""

    # Check that if we exceed max_redirects a MaxRedirectError is raised
    # Note: To have a stable test, the order of requirements matters. So this needs to
    # use a list or OrderedDict to guarantee that the_dcc==1.2 requirements are
    # processed before the_dcc_plugin_b which specifies the_dcc<1.2 forcing a redirect.
    requirements = OrderedDict(
        (
            ("the_dcc", Requirement("the_dcc")),
            ("the_dcc_plugin_b", Requirement("the_dcc_plugin_b==0.9")),
        )
    )

    solver = Solver(requirements, resolver)
    solver.max_redirects = 0
    with pytest.raises(MaxRedirectError, match="Redirect limit of 0 reached"):
        solver.resolve()



__all__ = ["__version__", "NotSet", "Resolver", "Site"]

from .utils import NotSet

# Note: Future imports depend on NotSet so it must be imported first
# isort: split

from .resolver import Resolver
from .site import Site
from .version import version as __version__



def gui():
    """Used to test entry point resolution. Raises an exception for ease of testing."""
    raise NotImplementedError("hab_test_entry_points.gui called successfully")


def gui_alt():
    """Used to test entry point resolution site modification. Raises an exception
    for ease of testing."""
    raise NotImplementedError("hab_test_entry_points.gui_alt called successfully")


def cfg_reduce_env(cfg):
    """Used to test that an entry point is called by raising an exception when
    called. See `tests/site/eps/README.md` for details."""
    raise NotImplementedError(
        "hab_test_entry_points.cfg_reduce_env called successfully"
    )


def cfg_reduce_finalize(cfg):
    """Used to test that an entry point is called by raising an exception when
    called. See `tests/site/eps/README.md` for details."""
    raise NotImplementedError(
        "hab_test_entry_points.cfg_reduce_finalize called successfully"
    )


def site_add_paths(site):
    """Add a couple of extra site paths to hab using `hab.site.add_paths` entry_point."""
    from pathlib import Path

    return [
        Path(__file__).parent / "site" / "eps" / "site_add_paths_a.json",
        Path(__file__).parent / "site" / "eps" / "site_add_paths_b.json",
    ]


def site_add_paths_a(site):
    """Add a couple of extra site paths to hab using `hab.site.add_paths` entry_point."""
    from pathlib import Path

    return [
        Path(__file__).parent / "site" / "eps" / "site_add_paths_c.json",
    ]


def site_finalize(site):
    """Used to test that an entry point is called by raising an exception when
    called. See `tests/site/eps/README.md` for details."""
    raise NotImplementedError("hab_test_entry_points.site_finalize called successfully")


def uri_validate_error(resolver, uri):
    """Used to test that an entry point is called by raising an exception when
    called. See `tests/site/eps/README.md` for details."""
    raise NotImplementedError(
        "hab_test_entry_points.uri_validate_error called successfully"
    )


def uri_validate_project_a(resolver, uri):
    """Used to test hab.uri.validate entry_point."""
    from hab.parsers import HabBase

    # Show how the URI can be modified by validation
    splits = uri.split(HabBase.separator)
    lower = splits[0].lower()
    if lower == "project_a" and splits[0] != lower:
        splits[0] = "project_a"
        uri = HabBase.separator.join(splits)
        return uri


def uri_validate_project_b(resolver, uri):
    """Used to test multiple hab.uri.validate entry_points."""
    from hab.parsers import HabBase

    # Raising an exception stops processing
    if uri == "raise-error":
        raise Exception('URI "raise-error" was used, raising an exception.')

    # Show how the URI can be modified by validation
    splits = uri.split(HabBase.separator)
    upper = splits[0].upper()
    if upper == "PROJECT_B" and splits[0] != upper:
        splits[0] = "PROJECT_B"
        uri = HabBase.separator.join(splits)
        return uri



import subprocess
import sys

from . import utils

try:
    CREATE_NO_WINDOW = subprocess.CREATE_NO_WINDOW
except AttributeError:
    # This constant comes from the WindowsAPI, but is not
    # defined in subprocess until python 3.7
    CREATE_NO_WINDOW = 0x08000000


class Launcher(subprocess.Popen):
    """Runs cmd using subprocess.Popen enabling stdout/err/in redirection.
    On windows prevents showing of command prompts when running with pythonw.

    Args:
        args (list or string): The command to be run by subprocess.
        **kwargs: Any keyword arguments are passed to subprocess.Popen. The
            standard defaults for **kwargs are modified to enable stdout/stderr
            capture to a single merged stdout stream. Also adds platform specific
            workarounds for stdin. Changes bufsize to 1 if possible to allow for
            per-line updating of output.
    """

    def __init__(self, args, **kwargs):
        # Change the defaults for subprocess.Popen to make it easier to capture
        # output from the launched subprocess.
        kwargs.setdefault("stdout", subprocess.PIPE)
        kwargs.setdefault("stderr", subprocess.STDOUT)
        kwargs.setdefault("universal_newlines", True)
        # In python 3.8+ changing bufsize when universal_newlines is False
        # causes a warning message.
        if kwargs["universal_newlines"] or kwargs.get("text") is True:
            kwargs.setdefault("bufsize", 1)

        if utils.Platform.name() == "windows" and "pythonw" in sys.executable.lower():
            # Windows 7 has an issue with making subprocesses
            # if the stdin handle is None, so pass the PIPE
            kwargs.setdefault("stdin", subprocess.PIPE)

            # If this is a pythonw process, because there is no current window
            # for stdout, any subprocesses will try to create a new window
            kwargs.setdefault("creationflags", CREATE_NO_WINDOW)

        super().__init__(args, **kwargs)



import logging
from copy import copy

from packaging.requirements import Requirement

from .errors import InvalidRequirementError, MaxRedirectError

logger = logging.getLogger(__name__)


class Solver(object):
    """Recursively check requirements into a flat set of requirements that satisfy the
    requirements and their requirements, etc.

    Args:
        requirements (list): The requirements to process and store in resolved.
        resolver (hab.Resolver): The Resolver used to define all distros and versions.
        forced (dict, optional): Forces this distro requirement replacing any resolved
            requirements. Using this may lead to configuring your environment
            incorrectly, use with caution.

    Attributes:
        invalid (dict, optional): If a recursive requirement makes a already resolved
            version invalid, that version is added to this list as an exclusive exclude.
    """

    def __init__(self, requirements, resolver, forced=None):
        self.forced = forced if forced else {}
        self.invalid = {}
        self.max_redirects = 2
        self.requirements = requirements
        self.resolver = resolver
        self.redirects_required = 0

    @classmethod
    def append_requirement(cls, requirements, req):
        """Combines any existing Requirement specifier with the new specifier.

        Args:
            requirements (dict): A dictionary of the requirements to modify.
            req (packaging.requirements.Requirement): The requirement to be combined
                with any existing requirement in requirements.

        Returns:
            The updated Requirement object that is also stored in requirements.
        """
        name = req.name
        if name in requirements:
            requirements[name].specifier &= req.specifier
        else:
            requirements[name] = copy(req)
        return requirements[name]

    def _resolve(
        self,
        requirements,
        resolved=None,
        processed=None,
        reported=None,
    ):
        """Recursively solve the provided requirements into a final list of requirements.

        Args:
            requirements (list): The requirements to process and store in resolved.
            resolved (dict): This dictionary is used to store the final resolved
                requirements matching the requested requirements.
            processed (set, optional): A set of `hab.parsers.DistroVersion`
                objects that have already been resolved. This prevents re-processing
                the same version over and over.

        Raises:
            ValueError: "Removing invalid version ..." This error indicates the need
                to re-run `_resolve`. A version added to the requirements that was later
                found to be no longer valid. A explicit `!=` version requirement has
                been added to `invalid` so calling `_resolve` again will prevent using
                that version for future dependency resolution.
        """

        # Set the default value for mutable objects
        if processed is None:
            processed = set()
        if reported is None:
            reported = set()
        if resolved is None:
            resolved = {}

        # Make sure we process the forced requirements
        if self.forced:
            # Note: Make sure to preserve the order of requirements, ie OrderedDict.
            # This is mostly to ensure that tests for solver errors are not flaky.
            requirements = requirements.copy()
            requirements.update(self.forced)

        if requirements:
            reqs = requirements.values()
        else:
            # Support the possibility that no configuration defines requirements
            reqs = []
        logger.debug("Requirements: {}".format(reqs))

        for req in reqs:
            name = req.name

            marker = req.marker
            if marker and not marker.evaluate():
                # If a marker is specified and its not valid for the current
                # system, skip this requirement.
                # https://packaging.pypa.io/en/stable/markers.html
                msg = f"Requirement ignored due to marker: {req}"
                if name in self.forced:
                    logger.critical(f"Forced {msg}")  # pragma: no cover
                else:
                    logger.warning(f"{msg}")
                continue

            if name in self.forced:
                if name in reported:
                    # Once we have processed this requirement, there is no need to
                    # re-process it, or log the warning again.
                    continue
                # If a distro requirement is forced, ignore the requested
                # requirement and instead resolve the forced requirement.
                req = self.forced[name]
                # This option should only be used for development and testing
                # always show a warning if its used.
                logger.warning("Forced Requirement: {}".format(req))
                reported.add(name)

            # Update the requirement to match all current requirements
            req = self.append_requirement(resolved, req)
            if name in self.invalid:
                # TODO: build the correct not specifier
                invalid = self.invalid[name]
                logger.debug("Adding invalid specifier: {}".format(invalid))
                req = req.specifier & invalid.specifier

            logger.debug("Checking requirement: {}".format(req))

            # Attempt to find a version, raises a exception if no version was found
            try:
                dist = self.resolver.distros[name]
            except KeyError:
                raise InvalidRequirementError(
                    f"Unable to find a distro for requirement: {req}"
                ) from None
            version = dist.latest_version(req)
            logger.debug("Found Version: {}".format(version.name))

            if version.distros and version not in processed:
                # Check if updated requirements have forced us to re-evaluate
                # our requirements.
                for v in processed:
                    if v.distro_name == version.distro_name:
                        invalid = Requirement("{}!={}".format(v.distro_name, v.version))
                        self.append_requirement(self.invalid, invalid)
                        raise ValueError(
                            "Removing invalid version {}".format(version.name)
                        )

                processed.add(version)
                self._resolve(version.distros, resolved, processed, reported)

        return resolved

    def resolve(self):
        """Recursively solve the provided requirements into a final list of requirements.

        Args:
            requirements (list): The requirements to process and store in resolved.
            resolved (dict): This dictionary is used to store the final resolved
                requirements matching the requested requirements.
            processed (set, optional): A set of `hab.parsers.DistroVersion`
                objects that have already been resolved. This prevents re-processing
                the same version over and over.

        Raises:
            MaxRedirectError: Redirect limit reached, unable to resolve the requested
                requirements.
        """

        self.redirects_required = 0
        logger.info("Resolving requirements: {}".format(self.requirements))
        while True:
            logger.info(
                "Attempt {} at resolving requirements".format(
                    self.redirects_required + 1
                )
            )
            try:
                return self._resolve(self.requirements)
            except ValueError as error:
                logger.info(str(error))
                self.redirects_required += 1
                if self.redirects_required >= self.max_redirects:
                    raise MaxRedirectError(
                        "Redirect limit of {} reached".format(self.max_redirects)
                    ) from None

    @classmethod
    def simplify_requirements(cls, requirements):
        """Convert various requirement formats to the expected dict format.

        Ultimately only str or Requirement objects can be passed to this function, but
        it can flatten lists, tuples and dictionaries into a single dictionary where
        the key is the simple requirement name and the value is a Requirement object.

        if requirements is an dictionary, both the key and value are added to the
        requirements. The value can be a list of additional requirements to be included.

        If a requirement is defined more than once, append_requirement will be called
        to merge the specifiers together into a single requirement per name.

        Returns:
            dict: A dictionary with requirement names as keys and
                ``packaging.requirements.Requirement`` objects with all requested
                specifiers included.
        """
        ret = requirements

        if isinstance(requirements, (list, tuple)):
            ret = {}
            for req in requirements:
                if not isinstance(req, Requirement):
                    req = Requirement(req)
                cls.append_requirement(ret, req)

        elif isinstance(requirements, dict):
            ret = {}
            for key, value in requirements.items():
                if isinstance(value, list):
                    # If value is a list, add all of its values
                    for val in cls.simplify_requirements(value).values():
                        cls.append_requirement(ret, val)
                elif value:
                    if not isinstance(value, Requirement):
                        value = Requirement(value)
                    cls.append_requirement(ret, value)

                # Add the key as a requirement even if an value was specified
                if key:
                    if not isinstance(key, Requirement):
                        key = Requirement(key)
                    cls.append_requirement(ret, key)

        return ret



""" Enables support for calling the hab cli using `python -m hab`
"""


import sys

import hab.cli

if __name__ == "__main__":
    # prog_name prevents __main__.py from being shown as the command name in the help
    # text. We don't know the exact command the user passed so we provide a generic
    # `python -m hab` command.
    sys.exit(hab.cli.cli(prog_name="python -m hab"))



import functools
import os
import re
import site
import subprocess
import sys

import pytest

from hab import Resolver, Site, utils
from hab.errors import HabError


class Topen(subprocess.Popen):
    """A custom subclass of Popen."""


def missing_annotations_hack(function):
    """Decorator that works around a missing annotations module in python 3.6.

    Allows for calling subprocess calls in python 3.6 that would normally fail
    with a `SyntaxError: future feature annotations is not defined` exception.

    Works by temporarily removing the `_virtualenv.pth` file in the venv's
    site-packages.

    TODO: Figure out a better method until we can drop CentOS requirement.
    """

    @functools.wraps(function)
    def new_function(*args, **kwargs):
        if sys.version_info.minor != 6:
            return function(*args, **kwargs)

        site_packages = site.getsitepackages()
        for path in site_packages:
            pth = os.path.join(path, "_virtualenv.pth")
            if os.path.exists(pth):
                os.rename(pth, f"{pth}.bak")
        try:
            ret = function(*args, **kwargs)
        finally:
            for path in site_packages:
                pth = os.path.join(path, "_virtualenv.pth.bak")
                if os.path.exists(pth):
                    os.rename(pth, pth[:-3])
        return ret

    return new_function


def test_launch(resolver):
    """Check the Config.launch method."""
    cfg = resolver.resolve("app/aliased/mod")
    proc = cfg.launch("global", args=None, blocking=True)

    # Enabling mod_std sends all exception text to stdout
    assert proc.output_stderr is None
    check = [
        " Env Vars: ".center(80, "-"),
        "ALIASED_GLOBAL_A: ['Local Mod A', 'Local A Prepend', 'Global A', 'Local A Append']",
        "ALIASED_GLOBAL_B: ['Global B']",
        "ALIASED_GLOBAL_C: ['Local C Set']",
        "ALIASED_GLOBAL_D: <UNSET>",
        "ALIASED_GLOBAL_E: <UNSET>",
        "ALIASED_LOCAL: <UNSET>",
        "CONFIG_DEFINED: ['config_variable']",
        "",
        " PATH env var ".center(80, "-"),
    ]

    assert "\n".join(check) in proc.output_stdout


@missing_annotations_hack
def test_launch_str(resolver):
    cfg = resolver.resolve("app/aliased/mod")

    # Check that args are passed to the subprocess
    # Check that if "cmd" is a str, it is converted to a list and extended with args
    args = ["-c", 'print("success")']
    proc = cfg.launch("as_str", args=args, blocking=True)
    assert proc.output_stdout.strip() == "success"

    # Check that passing env is respected
    env = dict(os.environ)
    var_name = "TEST_ADDED_VARIABLE"
    var_value = "Test variable"
    assert var_name not in env
    env[var_name] = var_value

    # Check that Popen kwargs can be passed through launch, including env.
    args = [
        "-c",
        'import os;assert os.environ["{}"] != "{}"'.format(var_name, var_value),
    ]
    proc = cfg.launch("as_str", args=args, blocking=True, env=env)


@pytest.mark.skipif(sys.platform != "win32", reason="only applies on windows")
def test_pythonw(monkeypatch, resolver):
    """Check that sys.stdin is set if using pythonw."""
    cfg = resolver.resolve("app/aliased/mod")

    # If not using pythonw.exe proc.stdin is not routed to PIPE
    proc = cfg.launch("global", args=None, blocking=True)
    assert proc.stdin is None

    # If using pythonw.exe, proc.stdin is routed to PIPE
    pyw = re.sub("python.exe", "pythonw.exe", sys.executable, flags=re.I)
    monkeypatch.setattr(sys, "executable", pyw)
    proc = cfg.launch("global", args=None, blocking=True)
    assert proc.stdin is not None


def test_invalid_alias(resolver):
    cfg = resolver.resolve("app/aliased/mod")
    with pytest.raises(HabError, match='"not-a-alias" is not a valid alias name'):
        cfg.launch("not-a-alias")

    # Remove the "cmd" value to test an invalid configuration
    alias = cfg.frozen_data["aliases"][utils.Platform.name()]["global"]
    del alias["cmd"]

    with pytest.raises(HabError, match='Alias "global" does not have "cmd" defined'):
        cfg.launch("global")


def test_cls_no_entry_point(resolver):
    """Check that if no entry point is defined, `hab.launcher.Launcher` is
    used to launch the alias.
    """
    entry_points = resolver.site.entry_points_for_group("hab.launch_cls")
    assert len(entry_points) == 0

    cfg = resolver.resolve("app/aliased/mod")
    proc = cfg.launch("global", blocking=True)

    from hab.launcher import Launcher

    assert type(proc) is Launcher

    # Check that specifying cls, overrides default
    proc = cfg.launch("global", blocking=True, cls=Topen)
    assert type(proc) is Topen


def test_cls_entry_point(config_root):
    """Check that if an entry point is defined, it is used unless overridden."""
    site = Site(
        [config_root / "site/site_entry_point_a.json", config_root / "site_main.json"]
    )
    entry_points = site.entry_points_for_group("hab.launch_cls")
    assert len(entry_points) == 1
    # Test that the `test-gui` `hab.cli` entry point is handled correctly
    ep = entry_points[0]
    assert ep.name == "subprocess"
    assert ep.group == "hab.launch_cls"
    assert ep.value == "subprocess:Popen"

    resolver = Resolver(site=site)
    cfg = resolver.resolve("app/aliased/mod")

    # Check that entry_point site config is respected
    proc = cfg.launch("global", blocking=True)
    assert type(proc) is subprocess.Popen

    # Check that specifying cls, overrides site config
    proc = cfg.launch("global", blocking=True, cls=Topen)
    assert type(proc) is Topen


def test_alias_entry_point(config_root):
    """Check that if an entry point is defined on a complex alias, it is used."""
    site = Site(
        [config_root / "site/site_entry_point_a.json", config_root / "site_main.json"]
    )

    resolver = Resolver(site=site)
    cfg = resolver.resolve("app/aliased/mod")

    # NOTE: We need to compare the name of the classes because they are separate
    # imports that don't compare equal using `is`.

    # Check that entry_point site config is respected
    proc = cfg.launch("global", blocking=True)
    assert type(proc).__name__ == "Popen"

    # Check that if the complex alias specifies hab.launch_cls, it is used instead
    # of the site defined or default class.
    alias = cfg.frozen_data["aliases"][utils.Platform.name()]["global"]
    alias["hab.launch_cls"] = {"subprocess": "tests.test_launch:Topen"}
    proc = cfg.launch("global", blocking=True)
    assert type(proc).__name__ == "Topen"



# __all__ = ["Resolver"]

import glob
import logging

import anytree
from packaging.requirements import Requirement

from . import utils
from .errors import _IgnoredVersionError
from .parsers import Config, DistroVersion, HabBase
from .site import Site
from .solvers import Solver
from .user_prefs import UserPrefs

logger = logging.getLogger(__name__)


class Resolver(object):
    """Used to resolve a hab environment setup and apply it to the current environment.

    Args:
        site (hab.Site, optional): The site configuration. if not provided, uses
            the ``HAB_PATHS`` environment variable to load the configuration.
        prereleases (bool, optional): When resolving distro versions, should
            pre-releases be included in the latest version. If not specified uses the
            value specified in site for the ``"prereleases"`` value.
        forced_requirements (list, optional): A list of additional version requirements
            to respect even if they are not specified in a config. This is provided for
            ease of hab package development and should not be used in production.
    """

    def __init__(
        self,
        site=None,
        prereleases=None,
        forced_requirements=None,
        target="hab",
    ):
        if site is None:
            site = Site()
        self.site = site

        if prereleases is None:
            prereleases = self.site.get("prereleases", False)
        self.prereleases = prereleases

        # Variables used to filter various outputs. See `hab.utils.verbosity_filter`
        # with context for more details.
        self._verbosity_target = target
        # A value of None indicates that nothing should be hidden, otherwise only
        # show the item if its min_verbosity is <= this number
        self._verbosity_value = None

        if forced_requirements:
            self.forced_requirements = Solver.simplify_requirements(forced_requirements)
        else:
            self.forced_requirements = {}

        self.config_paths = utils.Platform.expand_paths(
            self.site.get("config_paths", [])
        )
        self.distro_paths = utils.Platform.expand_paths(
            self.site.get("distro_paths", [])
        )
        logger.debug("config_paths: {}".format(self.config_paths))
        logger.debug("distro_paths: {}".format(self.distro_paths))

        self._configs = None
        self._distros = None
        self.ignored = self.site["ignored_distros"]

        # If true, then all scripts are printed instead of being written to disk
        # to allow for quick debugging of the underlying shell scripts driving hab
        # If the files are not written to disk, the calling shell scripts will
        # simply exit instead of running the requested command.
        self.dump_scripts = False

    def clear_caches(self):
        """Clears cached resolved data so it is re-generated on next use."""
        logger.debug("Resolver cache cleared.")
        self._configs = None
        self._distros = None

    def closest_config(self, path, default=False):
        """Returns the most specific leaf or the tree root matching path. Ignoring any
        path names that don't exist in self.configs.

        Args:
            path (str): A config path relative to the root of the tree.
            default (bool, optional): If True, search the default tree instead of the
                tree specified by the first name in path. The leaf nodes do not need to
                match names exactly, it will pick a default leaf that starts with the
                most common characters. Ie if path is `:project_a:Sc001` it would match
                `:default:Sc0` not `:default:Sc01`.
        """
        if not path.startswith(HabBase.separator):
            path = "".join((HabBase.separator, path))
        # Anytree<2.9.0 had a bug when resolving URI's that end in a slash like
        # `app/` would cause a IndexError. This ensure that older versions work
        path = path.rstrip("/")

        if default:
            node_names = path.split(HabBase.separator)
            current = self.configs["default"]
            # Skip the root and project name it won't match default
            for node_name in node_names[2:]:
                # Find the node that starts with the longest match
                matches = sorted(
                    [c for c in current.children if node_name.startswith(c.name)],
                    key=lambda i: i.name,
                    reverse=True,
                )
                if matches:
                    current = matches[0]
                else:
                    break
            return current

        # Handle the non-default lookup
        splits = path.split(HabBase.separator)
        # Find the forest to search for or return the default search
        root_name = splits[1 if path.startswith(HabBase.separator) else 0]
        if root_name not in self.configs:
            return self.closest_config(path, default=True)

        resolver = anytree.Resolver()

        try:
            # Workaround a zombie bug in anytree's glob. Where glob doesn't match
            # top level paths and will raise a IndexError incorrectly.
            # https://github.com/c0fec0de/anytree/issues/125
            if len(splits) > 2:
                items = resolver.glob(self.configs[root_name], path)
            else:
                return resolver.get(self.configs[root_name], path)
        except anytree.resolver.ResolverError as e:
            return e.node
        if items:
            return items[0]
        # TODO: If the anytree bug gets fixed we should start hitting this line.
        # Until then exclude it from the completeness check.
        return self.configs[root_name]  # pragma: no cover

    @property
    def config_paths(self):
        return self._config_paths

    @config_paths.setter
    def config_paths(self, paths):
        # Convert string paths into a list
        if isinstance(paths, str):
            paths = utils.Platform.expand_paths(paths)

        self._config_paths = paths
        # Reset _configs so we re-generate them the next time they are requested
        self._configs = None

    @property
    def configs(self):
        """A dictionary of all configurations that have been parsed for this resolver"""
        if self._configs is None:
            self._configs = self.parse_configs(self.config_paths)
        return self._configs

    @property
    def distro_paths(self):
        return self._distro_paths

    @distro_paths.setter
    def distro_paths(self, paths):
        # Convert string paths into a list
        if isinstance(paths, str):
            paths = utils.Platform.expand_paths(paths)

        self._distro_paths = paths
        # Reset _distros so we re-generate them the next time they are requested
        self._distros = None

    @property
    def distros(self):
        """A list of all of the requested distros to resolve."""
        if self._distros is None:
            self._distros = self.parse_distros(self.distro_paths)
        return self._distros

    @classmethod
    def dump_forest(
        cls,
        forest,
        attr="uri",
        fmt="{pre}{attr}",
        style=None,
        indent="  ",
        truncate=None,
    ):
        """Yields the contents of a forest to a readable string or anytree objects.

        Args:
            forest: A dictionary of hab.parser objects to dump. Common values
                are `resolver.configs`, or `resolver.distros`.
            attr (str, optional): The name of the attribute to display for each node.
                If None is passed, the anytree object is returned un-modified.
            fmt (str, optional): str.format string to control the display of
                each node in the forest. Accepts (pre, attr) keys.
            style (anytree.render.AbstractStyle, optional): Controls how anytree
                renders the branch information. If not set, defaults to a custom
                style that intents all children(recursively) to the same depth.
            indent (str, optional): The string to use for indentation. Used only
                if style is None.
            truncate (int, optional): The maximum number of results to show for
                a given tree level. If there are more child nodes than twice this
                value, only include the first and last number of these results
                with a "..." placeholder in between. Disable by passing None.
        """

        def sort_forest(items):
            """Ensures consistent sorting of the forest leaf nodes"""
            ret = utils.natural_sort(items, key=lambda item: item.name)
            if truncate and len(ret) > truncate * 2:
                # Create a anytree node that can be rendered as "..."
                placeholder = HabBase._placeholder(forest, None)
                placeholder.name = "..."
                # replace the excess middle nodes with the placeholder object
                ret = ret[:truncate] + [placeholder] + ret[-truncate:]
            return ret

        limit_pre = False
        if style is None:
            style = anytree.render.AbstractStyle(indent, indent, indent)
            limit_pre = True

        for tree_name in utils.natural_sort(forest):
            for row in anytree.RenderTree(
                forest[tree_name], style=style, childiter=sort_forest
            ):
                cfg = row.node
                # Process inheritance for this config to ensure the correct value
                cfg._collect_values(cfg, ["min_verbosity"])

                # Check if this row should be shown based on verbosity settings.
                min_verbosity = {"min_verbosity": cfg.min_verbosity}
                if not cfg.check_min_verbosity(min_verbosity):
                    # TODO: If a parent is hidden but not a child, fix rendering
                    # so the child's indent is reduced correctly.
                    continue

                # Yield the anytree node if no attr was requested
                if attr is None:
                    yield row
                else:
                    # Otherwise yield the formatted text for the node
                    pre = row.pre
                    if limit_pre:
                        pre = row.pre[: len(indent)]
                    yield fmt.format(pre=pre, attr=getattr(cfg, attr))

    def find_distro(self, requirement):
        """Returns the DistroVersion matching the requirement or None"""
        if not isinstance(requirement, Requirement):
            requirement = Requirement(requirement)

        if requirement.name in self.distros:
            app = self.distros[requirement.name]
            return app.latest_version(requirement)

    def parse_configs(self, config_paths, forest=None):
        if forest is None:
            forest = {}
        for dirname in config_paths:
            for path in sorted(glob.glob(str(dirname / "*.json"))):
                Config(forest, self, path, root_paths=set((dirname,)))
        return forest

    def parse_distros(self, distro_paths, forest=None):
        if forest is None:
            forest = {}
        for dirname in distro_paths:
            for path in sorted(glob.glob(str(dirname / "*" / ".hab.json"))):
                try:
                    DistroVersion(forest, self, path, root_paths=set((dirname,)))
                except _IgnoredVersionError as error:
                    logger.debug(str(error))
        return forest

    def resolve(self, uri):
        """Find the closest configuration and reduce it into its final form."""
        uri = self.uri_validate(uri)
        context = self.closest_config(uri)
        return context.reduced(self, uri=uri)

    def resolve_requirements(self, requirements):
        """Recursively solve the provided requirements into a final list of requirements.

        Args:
            requirements (list): The requirements to resolve.

        Raises:
            MaxRedirectError: Redirect limit reached, unable to resolve the requested
                requirements.
        """

        solver = Solver(requirements, self, forced=self.forced_requirements)
        return solver.resolve()

    def uri_validate(self, uri):
        """Check for issues with the provided URI and possibly modify it.

        This runs all `hab.uri.validate` entry points specified by site. These should
        point to a callable that accepts `resolver` and `uri` kwargs. If an URI
        is not valid, then an exception should be raised. The callback can update
        the URI by returning a string.
        """
        # Run any configured entry_points before aliases are calculated
        for ep in self.site.entry_points_for_group("hab.uri.validate"):
            logger.debug(f"Running hab.uri.validate entry_point: {ep}")
            func = ep.load()
            ret = func(resolver=self, uri=uri)
            if ret:
                uri = ret
        return uri

    def user_prefs(self, load=False):
        """Returns the `hab.user_prefs.UserPrefs` object for this resolver.

        Args:
            load (bool, optional): If True, calls `UserPrefs.load()` before
                returning the UserPrefs object.
        """
        try:
            self._user_prefs
        except AttributeError:
            self._user_prefs = UserPrefs(self)

        if load:
            self._user_prefs.load()
        return self._user_prefs



import logging
import os

from .. import NotSet, utils
from ..errors import HabError
from .hab_base import HabBase
from .meta import hab_property

logger = logging.getLogger(__name__)


class Config(HabBase):
    """The configuration for a given URI that defines required distros and environment
    variables need to be loaded if this config is chosen. This does not resolve `NotSet`
    values, see `FlatConfig` for the final resolved values that are actually applied."""

    def __init__(self, *args, **kwargs):
        self._alias_mods = NotSet
        super().__init__(*args, **kwargs)

    @hab_property(process_order=120)
    def aliases(self):
        """Dict of the names and commands that need created to launch desired
        applications."""
        ret = self.frozen_data.get("aliases", {}).get(utils.Platform.name(), {})
        # Only return aliases if they are valid for the current verbosity
        return {k: v for k, v in ret.items() if self.check_min_verbosity(v)}

    # Note: 'alias_mods' needs to be processed before 'environment'
    @hab_property(verbosity=3, process_order=50)
    def alias_mods(self):
        """Dict of modifications that need to be made on aliases.
        These are used to modify the original configuration of an alias by another
        distro or config. This allows a plugin to add an environment variable to
        a specific alias even though the alias is defined by another distro/config.
        """
        return self._alias_mods

    @hab_property(verbosity=2)
    def inherits(self):
        return self.frozen_data.get("inherits", NotSet)

    @inherits.setter
    def inherits(self, inherits):
        self.frozen_data["inherits"] = inherits

    def launch(self, alias_name, args=None, blocking=False, cls=None, **kwargs):
        """Launches the requested alias using subprocess.Popen.

        Args:
            alias_name (str): The alias name to run.
            args (list): The command to be run by subprocess. This should be a list
                of each individual string argument. If a kwarg is being passed it
                should be passed as two items. ['--key', 'value'].
            blocking (bool or str, optional): Makes this method blocking by calling
                Popen.communicate. If a str value is used, is included in the call.
                The results of calling `proc.communicate` can be accessed from the
                `output_stdout` and `output_stderr` properties added to proc.
            cls (class, optional): A `subprocess.Popen` compatible class is
                initialized and used to run the alias. If not passed, then the
                site entry_point `hab.launch_cls` is used if defined. Otherwise the
                `hab.launcher.Launcher` class is used.
            **kwargs: Any keyword arguments are passed to subprocess.Popen. If on
                windows and using pythonw, prevents showing a command prompt.

        Returns:
            The created subprocess.Popen instance.
        """
        # Construct the command line arguments to execute
        if alias_name not in self.aliases:
            raise HabError(f'"{alias_name}" is not a valid alias name')
        alias = self.aliases[alias_name]

        # Get the subprocess.Popen like class to use to launch the alias
        if cls is None:
            # Use the entry_point if defined on the alias
            alias_cls = alias.get("hab.launch_cls")
            if alias_cls:
                alias_cls = {"hab.launch_cls": alias_cls}
                eps = self.resolver.site.entry_points_for_group(
                    "hab.launch_cls", entry_points=alias_cls
                )
            else:
                # Otherwise use the global definition from Site
                eps = self.resolver.site.entry_points_for_group("hab.launch_cls")

            if eps:
                cls = eps[0].load()
            else:
                # Default to subprocess.Popen if not defined elsewhere
                from hab.launcher import Launcher

                cls = Launcher

        try:
            cmd = alias["cmd"]
        except KeyError:
            raise HabError(
                f'Alias "{alias_name}" does not have "cmd" defined'
            ) from None
        if isinstance(cmd, str):
            cmd = [cmd]
        if args:
            cmd.extend(args)

        # Apply the hab global and alias environment variable changes
        if "env" in kwargs:
            env = kwargs["env"]
        else:
            env = dict(os.environ)
        self.update_environ(env, alias_name)
        kwargs["env"] = env

        # Launch the subprocess using the requested Popen subclass
        logger.info(f"Launching: {cmd}")
        proc = cls(cmd, **kwargs)

        if blocking:
            if blocking is True:
                blocking = None
            proc.output_stdout, proc.output_stderr = proc.communicate(input=blocking)

        return proc

    def load(self, filename):
        data = super().load(filename)
        self._alias_mods = data.get("alias_mods", NotSet)
        self.inherits = data.get("inherits", NotSet)
        return data

    @hab_property(verbosity=1, group=0)
    def uri(self):
        # Mark uri as a HabProperty so it is included in _properties
        return super().uri



import os
from collections import OrderedDict
from pathlib import Path

import anytree
import pytest
from packaging.requirements import Requirement

from hab import NotSet, Resolver, Site, utils
from hab.solvers import Solver


def test_environment_variables(config_root, helpers, monkeypatch):
    """Check that Resolver's init respects the environment variables it uses."""
    config_paths_env = utils.Platform.expand_paths(["a/config/path", "b/config/path"])
    distro_paths_env = utils.Platform.expand_paths(["a/distro/path", "b/distro/path"])
    config_paths_direct = utils.Platform.expand_paths(
        ["z/config/path", "zz/config/path"]
    )
    distro_paths_direct = utils.Platform.expand_paths(
        ["z/distro/path", "zz/distro/path"]
    )

    # Set the config environment variables
    monkeypatch.setenv(
        "HAB_PATHS", utils.Platform.collapse_paths([config_root / "site_env.json"])
    )

    # configured by the environment
    resolver_env = Resolver()
    # configured by passed arguments
    resolver_direct = Resolver(site=Site([config_root / "site_env_direct.json"]))

    # Check the environment configured resolver
    assert resolver_env.config_paths == config_paths_env
    assert resolver_env.distro_paths == distro_paths_env

    # Check the directly configured resolver
    assert resolver_direct.config_paths == config_paths_direct
    assert resolver_direct.distro_paths == distro_paths_direct

    # Check that we properly split string paths into a lists if provided
    resolver_env = Resolver(site=Site([config_root / "site_env.json"]))
    assert resolver_env.config_paths == config_paths_env
    assert resolver_env.distro_paths == distro_paths_env


def test_config(resolver):
    """Spot check a few of the parsed configs to ensure config works."""
    any_resolver = anytree.Resolver()
    default = resolver.configs["default"]
    assert default.name == "default"
    assert any_resolver.get(default, "/default/Sc1").name == "Sc1"
    assert any_resolver.get(default, "/default/Sc11").name == "Sc11"


@pytest.mark.parametrize(
    "path,result,reason",
    (
        ("project_a", "project_a", "Complete root path not found"),
        ("project_a/Sc001", "project_a/Sc001", "Complete secondary path not found"),
        (
            "project_a/Sc001/Animation",
            "project_a/Sc001/Animation",
            "Complete tertiary path not found",
        ),
        (
            "project_a/Sc001/Modeling",
            "project_a/Sc001",
            "Invalid tertiary path not fallen back",
        ),
        (
            "project_a/Sc999/Modeling",
            "project_a",
            "Invalid secondary path not fallen back",
        ),
        (
            "project_a/very/many/paths/resolved",
            "project_a",
            "Invalid n-length path not fallen back",
        ),
        # Default fallback
        ("project_z", "default", "Default root not returned for invalid root."),
        (
            "project_z/Sc001",
            "default",
            "Default root not returned if no matching secondary.",
        ),
        ("project_z/Sc101", "default/Sc1", "Default secondary not returned"),
        (
            "project_z/Sc110",
            "default/Sc11",
            "More specific default secondary not returned",
        ),
        # Leading/trailing separators
        ("/app", "app", "Leading slash not ignored"),
        ("app/", "app", "Trailing slash not sanitized correctly"),
        ("app/case/", "app", "Trailing slash not sanitized correctly"),
    ),
)
def test_closest_config(resolver, path, result, reason):
    """Test that closest_config returns the expected results."""
    assert resolver.closest_config(path).fullpath == result, reason


class TestDumpForest:
    """Test the dump_forest method on resolver"""

    all_uris = [
        "app",
        "  app/aliased",
        "  app/aliased/config",
        "  app/aliased/mod",
        "  app/aliased/mod/config",
        "  app/houdini",
        "  app/houdini/a",
        "  app/houdini/b",
        "  app/maya",
        "  app/maya/2020",
        "default",
        "  default/Sc1",
        "  default/Sc11",
        "not_set",
        "  not_set/child",
        "  not_set/distros",
        "  not_set/empty_lists",
        "  not_set/env1",
        "  not_set/env_path_hab_uri",
        "  not_set/env_path_set",
        "  not_set/env_path_unset",
        "  not_set/no_distros",
        "  not_set/no_env",
        "  not_set/os",
        "place-holder",
        "  place-holder/child",
        "  place-holder/inherits",
        "project_a",
        "  project_a/Sc001",
        "  project_a/Sc001/Animation",
        "  project_a/Sc001/Rigging",
        "verbosity",
        "  verbosity/hidden",
        "  verbosity/inherit",
        "  verbosity/inherit-no",
        "  verbosity/inherit-override",
    ]

    def test_uris_target_default(self, resolver):
        """Test dumping of configs using the default target `hab`. This also
        falls back to global setting."""
        check = list(self.all_uris)
        # Verbosity filter disabled, show all results
        assert resolver._verbosity_target == "hab"
        assert resolver._verbosity_value is None
        result = list(resolver.dump_forest(resolver.configs))
        assert result == check

        # Verbosity filter most verbose
        with utils.verbosity_filter(resolver, verbosity=2):
            result = list(resolver.dump_forest(resolver.configs))
        check.remove("  verbosity/hidden")
        assert result == check

        # Verbosity filter less verbose
        with utils.verbosity_filter(resolver, verbosity=1):
            result = list(resolver.dump_forest(resolver.configs))
        check.remove("  app/maya")
        check.remove("  app/maya/2020")
        check.remove("verbosity")
        check.remove("  verbosity/inherit")
        assert result == check

        # Verbosity filter least verbose
        with utils.verbosity_filter(resolver, verbosity=0):
            result = list(resolver.dump_forest(resolver.configs))
        check.remove("  verbosity/inherit-override")
        assert result == check

    def test_uris_objs(self, resolver):
        # If attr is "uri" then the uri as a string is returned.
        result = list(resolver.dump_forest(resolver.configs, attr="uri"))
        assert result == self.all_uris
        # If None is passed to attr, then the anytree object is returned
        for i, row in enumerate(resolver.dump_forest(resolver.configs, attr=None)):
            check = self.all_uris[i].strip()
            # Get the uri from the node
            assert check == row.node.uri

    def test_uris_target_hab_gui(self, resolver):
        """Test dumping of configs using the non-default target `hab-gui`. This
        also falls back to global setting."""
        check = list(self.all_uris)
        # Verbosity filter disabled, show all results
        assert resolver._verbosity_target == "hab"
        assert resolver._verbosity_value is None
        with utils.verbosity_filter(resolver, verbosity=None, target="hab-gui"):
            result = list(resolver.dump_forest(resolver.configs))
        assert result == check

        # Verbosity filter most verbose
        with utils.verbosity_filter(resolver, verbosity=2, target="hab-gui"):
            result = list(resolver.dump_forest(resolver.configs))
        assert result == check

        # Verbosity filter less verbose
        with utils.verbosity_filter(resolver, verbosity=1, target="hab-gui"):
            result = list(resolver.dump_forest(resolver.configs))
        check.remove("  verbosity/hidden")
        assert result == check

        # Verbosity filter least verbose
        with utils.verbosity_filter(resolver, verbosity=0, target="hab-gui"):
            result = list(resolver.dump_forest(resolver.configs))
        check.remove("  app/maya")
        check.remove("  app/maya/2020")
        check.remove("verbosity")
        check.remove("  verbosity/inherit")
        check.remove("  verbosity/inherit-override")
        assert result == check

    def test_distros(self, resolver):
        """Test dumping distros using name attr"""
        result = list(resolver.dump_forest(resolver.distros, attr="name"))
        check = [
            "aliased",
            "  aliased==2.0",
            "aliased_mod",
            "  aliased_mod==1.0",
            "aliased_verbosity",
            "  aliased_verbosity==1.0",
            "all_settings",
            "  all_settings==0.1.0.dev1",
            "houdini18.5",
            "  houdini18.5==18.5.351",
            "houdini19.5",
            "  houdini19.5==19.5.493",
            "maya2020",
            "  maya2020==2020.0",
            "  maya2020==2020.1",
            "the_dcc",
            "  the_dcc==1.0",
            "  the_dcc==1.1",
            "  the_dcc==1.2",
            "the_dcc_plugin_a",
            "  the_dcc_plugin_a==0.9",
            "  the_dcc_plugin_a==1.0",
            "  the_dcc_plugin_a==1.1",
            "the_dcc_plugin_b",
            "  the_dcc_plugin_b==0.9",
            "  the_dcc_plugin_b==1.0",
            "  the_dcc_plugin_b==1.1",
            "the_dcc_plugin_c",
            "  the_dcc_plugin_c==0.9",
            "  the_dcc_plugin_c==1.0",
            "  the_dcc_plugin_c==1.1",
            "the_dcc_plugin_d",
            "  the_dcc_plugin_d==0.9",
            "  the_dcc_plugin_d==1.0",
            "  the_dcc_plugin_d==1.1",
            "the_dcc_plugin_e",
            "  the_dcc_plugin_e==0.9",
            "  the_dcc_plugin_e==1.0",
            "  the_dcc_plugin_e==1.1",
        ]
        assert result == check

    def test_distros_truncate(self, resolver):
        """Test truncate feature by dumping distros"""
        result = list(resolver.dump_forest(resolver.distros, attr="name", truncate=1))
        check = [
            "aliased",
            "  aliased==2.0",
            "aliased_mod",
            "  aliased_mod==1.0",
            "aliased_verbosity",
            "  aliased_verbosity==1.0",
            "all_settings",
            "  all_settings==0.1.0.dev1",
            "houdini18.5",
            "  houdini18.5==18.5.351",
            "houdini19.5",
            "  houdini19.5==19.5.493",
            "maya2020",
            "  maya2020==2020.0",
            "  maya2020==2020.1",
            "the_dcc",
            "  the_dcc==1.0",
            "  ...",
            "  the_dcc==1.2",
            "the_dcc_plugin_a",
            "  the_dcc_plugin_a==0.9",
            "  ...",
            "  the_dcc_plugin_a==1.1",
            "the_dcc_plugin_b",
            "  the_dcc_plugin_b==0.9",
            "  ...",
            "  the_dcc_plugin_b==1.1",
            "the_dcc_plugin_c",
            "  the_dcc_plugin_c==0.9",
            "  ...",
            "  the_dcc_plugin_c==1.1",
            "the_dcc_plugin_d",
            "  the_dcc_plugin_d==0.9",
            "  ...",
            "  the_dcc_plugin_d==1.1",
            "the_dcc_plugin_e",
            "  the_dcc_plugin_e==0.9",
            "  ...",
            "  the_dcc_plugin_e==1.1",
        ]
        assert result == check


def test_reduced(resolver, helpers):
    """Check that NotSet is used if no value is provided."""
    cfg = resolver.closest_config("not_set")
    check = ["aliased", "maya2020"]
    helpers.assert_requirements_equal(cfg.distros, check)
    assert cfg.environment_config == NotSet
    assert cfg.inherits is False
    assert cfg.name == "not_set"

    cfg = resolver.closest_config("not_set/child")
    # Not set on the child so should be NotSet(ie doesn't inherit from parent)
    assert cfg.distros == NotSet
    # Settings defined on the child
    config_check = {
        "set": {"TEST": "case", "FMT_FOR_OS": "a{;}b;c:{PATH!e}{;}d"},
        "unset": ["UNSET_VARIABLE"],
    }
    assert cfg.environment_config == config_check
    assert cfg.inherits is True
    assert cfg.name == "child"

    # Verify that a flattened config properly inherits values
    reduced = cfg.reduced(resolver)
    # Inherited from the parent
    helpers.assert_requirements_equal(reduced.distros, check)
    # Values defined on the child are preserved
    assert reduced.environment_config == config_check
    assert reduced.inherits is True
    assert reduced.name == "child"
    assert reduced.uri == "not_set/child"

    # Verify that uri is handled correctly
    uri = "not_set/child/test"
    reduced = cfg.reduced(resolver, uri=uri)
    assert reduced._uri == uri
    assert reduced.uri == uri
    cfg = resolver.closest_config(uri)
    assert cfg._uri is NotSet
    assert cfg.uri == "not_set/child"


def test_resolve_requirements_simple(resolver):
    requirements = {
        "the_dcc": Requirement("the_dcc"),
    }

    # A simple resolve with no recalculations
    resolved = resolver.resolve_requirements(requirements)

    assert len(resolved) == 5
    assert str(resolved["the_dcc"]) == "the_dcc"
    # required by the_dcc==1.2 distro
    assert str(resolved["the_dcc_plugin_a"]) == "the_dcc_plugin_a>=1.0"
    assert str(resolved["the_dcc_plugin_b"]) == "the_dcc_plugin_b>=0.9"
    assert str(resolved["the_dcc_plugin_e"]) == "the_dcc_plugin_e<2.0"
    # required by the_dcc_plugin_a distro
    assert str(resolved["the_dcc_plugin_d"]) == "the_dcc_plugin_d"

    # Check the versions
    assert resolver.find_distro(resolved["the_dcc"]).name == "the_dcc==1.2"
    assert (
        resolver.find_distro(resolved["the_dcc_plugin_a"]).name
        == "the_dcc_plugin_a==1.1"
    )
    assert (
        resolver.find_distro(resolved["the_dcc_plugin_b"]).name
        == "the_dcc_plugin_b==1.1"
    )
    assert (
        resolver.find_distro(resolved["the_dcc_plugin_d"]).name
        == "the_dcc_plugin_d==1.1"
    )
    assert (
        resolver.find_distro(resolved["the_dcc_plugin_e"]).name
        == "the_dcc_plugin_e==1.1"
    )

    # Check that we can pass a string not a Requirement object to find_distro
    assert resolver.find_distro("the_dcc==1.2").name == "the_dcc==1.2"


def test_resolve_requirements_recalculate(resolver):
    """The first pick "the_dcc==1.2" gets discarded by plugin_b. Make sure the correct
    distros are picked.
    """

    # Resolve requires re-calculating
    # Note: To have a stable test, the order of requirements matters. Use an OrderedDict
    requirements = OrderedDict(
        (
            ("the_dcc", Requirement("the_dcc")),
            ("the_dcc_plugin_b", Requirement("the_dcc_plugin_b==0.9")),
        )
    )

    # Use the underlying Solver so we have access to debug resolve_requirements obscures
    solver = Solver(requirements, resolver)
    resolved = solver.resolve()

    # Check that we had to recalculate the resolve at least one time.
    assert solver.redirects_required == 1
    # Check that the_dcc 1.2 had to be ignored, triggering the recalculate
    assert len(solver.invalid) == 1
    assert str(solver.invalid["the_dcc"]) == "the_dcc!=1.2"

    # Check that the resolve is correct
    assert len(resolved) == 5
    assert str(resolved["the_dcc"]) == "the_dcc<1.2"
    # required by the_dcc==1.1 distro
    assert str(resolved["the_dcc_plugin_a"]) == "the_dcc_plugin_a>=1.0"
    assert str(resolved["the_dcc_plugin_b"]) == "the_dcc_plugin_b==0.9"
    assert str(resolved["the_dcc_plugin_e"]) == "the_dcc_plugin_e<2.0"
    # required by the_dcc_plugin_a distro
    assert str(resolved["the_dcc_plugin_d"]) == "the_dcc_plugin_d"

    # Check the versions
    assert resolver.find_distro(resolved["the_dcc"]).name == "the_dcc==1.1"
    assert (
        resolver.find_distro(resolved["the_dcc_plugin_a"]).name
        == "the_dcc_plugin_a==1.1"
    )
    assert (
        resolver.find_distro(resolved["the_dcc_plugin_b"]).name
        == "the_dcc_plugin_b==0.9"
    )
    assert (
        resolver.find_distro(resolved["the_dcc_plugin_d"]).name
        == "the_dcc_plugin_d==1.1"
    )
    assert (
        resolver.find_distro(resolved["the_dcc_plugin_e"]).name
        == "the_dcc_plugin_e==1.1"
    )


@pytest.mark.parametrize(
    "platform,marker",
    (
        ("windows", "Windows"),
        ("linux", "Linux"),
        ("osx", "Darwin"),
    ),
)
def test_resolve_requirements_markers(resolver, platform, marker):
    """Test that platform_system for current host is included or excluded correctly.

    The packaging marker library isn't setup to allow for testing on other
    platforms, so these tests need to pass on all platforms, if running the test
    on this platform, the dependency is included, otherwise it should be ignored.
    """
    check = [
        "the_dcc",
        "the_dcc_plugin_a",
        "the_dcc_plugin_b",
        "the_dcc_plugin_d",
        "the_dcc_plugin_e",
    ]
    # This requirement is only included if running on the target platform
    if utils.Platform.name() == platform:
        check.append("the_dcc_plugin_c")

    # Build requirements utilizing the platform marker.
    requirements = {
        "the_dcc": Requirement("the_dcc"),
        # the_dcc_plugin_c is only included if the current platform matches that.
        "the_dcc_plugin_c": Requirement(
            f"the_dcc_plugin_c;platform_system=='{marker}'"
        ),
    }

    ret = resolver.resolve_requirements(requirements)
    assert set(ret.keys()) == set(check)


@pytest.mark.parametrize(
    "forced,check,check_versions",
    (
        # No forced items
        (
            None,
            ["the_dcc_plugin_a", "the_dcc_plugin_d", "the_dcc_plugin_e<1.0,<2.0"],
            [],
        ),
        # Force
        (
            {
                # Adds a completely new requirement not specified in the config
                "the_dcc_plugin_c": Requirement("the_dcc_plugin_c"),
                # Forces the requirement to a invalid version for the config
                "the_dcc_plugin_e": Requirement("the_dcc_plugin_e==1.1"),
            },
            [
                "the_dcc_plugin_a",
                "the_dcc_plugin_c",
                "the_dcc_plugin_d",
                "the_dcc_plugin_e==1.1",
            ],
            ["the_dcc_plugin_c==1.1", "the_dcc_plugin_e==1.1"],
        ),
    ),
)
def test_forced_requirements(resolver, helpers, forced, check, check_versions):
    requirements = {
        # plugin_a adds an extra dependency outside of the requirements or forced
        "the_dcc_plugin_a": Requirement("the_dcc_plugin_a"),
        "the_dcc_plugin_e": Requirement("the_dcc_plugin_e<1.0"),
    }

    # Check that forced_requirement's are included in the resolved requirements
    resolver_forced = Resolver(
        site=resolver.site,
        forced_requirements=forced,
    )
    resolved = resolver_forced.resolve_requirements(requirements)
    helpers.assert_requirements_equal(resolved, check)

    # Check that forced_requirements work if the config defines zero distros
    cfg = resolver_forced.resolve("not_set/no_distros")
    versions = cfg.versions
    assert len(versions) == len(check_versions)
    for i, v in enumerate(versions):
        assert v.name == check_versions[i]


@pytest.mark.parametrize(
    "value,check",
    (
        ("test_string", [Path("test_string")]),
        (f"one{os.pathsep}two", [Path("one"), Path("two")]),
    ),
)
def test_path_expansion(resolver, value, check):
    # Both of these properties end up calling utils.expand_path, so
    # this doubles as a test for that edge case.
    resolver.config_paths = value
    assert resolver.config_paths == check

    resolver.distro_paths = value
    assert resolver.distro_paths == check

    # Check that collapse_paths also works as expected
    assert utils.Platform.collapse_paths("test_string") == str("test_string")


class TestPlatform:
    def test_collapse_paths(self):
        # NOTE: The `ext=".sh", key="PATH"` checks are covering the special case
        # that exists for PATH when using cygwin that doesn't apply to other
        # environment variables. See `utils.WinPlatform.collapse_paths` for details.

        # Passing strings
        arg = "test case"
        assert utils.LinuxPlatform.collapse_paths(arg) == "test case"
        assert utils.OsxPlatform.collapse_paths(arg) == "test case"
        assert utils.WinPlatform.collapse_paths(arg) == "test case"
        assert utils.WinPlatform.collapse_paths(arg, ext="") == "test case"
        assert utils.WinPlatform.collapse_paths(arg, ext=".sh") == "test case"
        assert (
            utils.WinPlatform.collapse_paths(arg, ext=".sh", key="PATH") == "test case"
        )

        # Passing lists that are not paths
        arg = ["test", "case"]
        assert utils.LinuxPlatform.collapse_paths(arg) == "test:case"
        assert utils.OsxPlatform.collapse_paths(arg) == "test:case"
        assert utils.WinPlatform.collapse_paths(arg) == "test;case"
        assert utils.WinPlatform.collapse_paths(arg, ext="") == "test;case"
        assert utils.WinPlatform.collapse_paths(arg, ext=".sh") == "test;case"
        assert (
            utils.WinPlatform.collapse_paths(arg, ext=".sh", key="PATH") == "test:case"
        )

        # Passing lists that are windows paths
        arg = ["c:\\test", "C:\\case"]
        assert utils.LinuxPlatform.collapse_paths(arg) == "c:\\test:C:\\case"
        assert utils.OsxPlatform.collapse_paths(arg) == "c:\\test:C:\\case"
        assert utils.WinPlatform.collapse_paths(arg) == "c:\\test;C:\\case"
        assert utils.WinPlatform.collapse_paths(arg, ext="") == "c:\\test;C:\\case"
        assert utils.WinPlatform.collapse_paths(arg, ext=".sh") == "c:\\test;C:\\case"
        assert (
            utils.WinPlatform.collapse_paths(arg, ext=".sh", key="PATH")
            == "/c/test:/C/case"
        )

        # Passing lists that are linux paths
        arg = ["/test", "/case"]
        assert utils.LinuxPlatform.collapse_paths(arg) == "/test:/case"
        assert utils.OsxPlatform.collapse_paths(arg) == "/test:/case"
        assert utils.WinPlatform.collapse_paths(arg) == "/test;/case"
        assert utils.WinPlatform.collapse_paths(arg, ext="") == "/test;/case"
        assert utils.WinPlatform.collapse_paths(arg, ext=".sh") == "/test;/case"
        assert (
            utils.WinPlatform.collapse_paths(arg, ext=".sh", key="PATH")
            == "/test:/case"
        )

    def test_pathsep(self):
        assert utils.LinuxPlatform.pathsep() == ":"
        assert utils.OsxPlatform.pathsep() == ":"
        assert utils.WinPlatform.pathsep() == ";"
        # Ext is not ignored for windows
        assert utils.WinPlatform.pathsep(ext="") == ";"
        assert utils.WinPlatform.pathsep(ext=".sh") == ";"
        assert utils.WinPlatform.pathsep(ext=".sh", key="PATH") == ":"


def test_cygpath():
    # Check space handling
    assert utils.cygpath("test case") == "test case"
    assert utils.cygpath("test  case") == "test  case"
    assert utils.cygpath("test case", spaces=True) == "test\\ case"
    assert utils.cygpath("test  case", spaces=True) == "test\\ \\ case"
    assert utils.cygpath("more test  case", spaces=True) == "more\\ test\\ \\ case"
    assert utils.cygpath("\\test case\\") == "/test case/"
    assert utils.cygpath("\\test case\\", spaces=True) == "/test\\ case/"

    # Check converting back-slashes to forward-slashes and drive letter handling
    assert utils.cygpath("c:\\test\\case") == "/c/test/case"
    assert utils.cygpath("C:\\test\\case") == "/C/test/case"
    assert utils.cygpath("\\\\server\\share\\dir") == "//server/share/dir"
    assert utils.cygpath("//server/share/dir") == "//server/share/dir"
    assert utils.cygpath("\\\\server/share/dir") == "//server/share/dir"
    assert utils.cygpath("/C/test/case") == "/C/test/case"

    # Check paths with spaces
    def cyg_space(path):
        return utils.cygpath(path, spaces=True)

    assert utils.cygpath("c:\\test\\spaces  are bad") == "/c/test/spaces  are bad"
    assert cyg_space("c:\\test\\spaces  are bad") == "/c/test/spaces\\ \\ are\\ bad"
    assert (
        utils.cygpath("//server/share/spaces  are bad")
        == "//server/share/spaces  are bad"
    )
    assert (
        cyg_space("//server/share/spaces  are bad")
        == "//server/share/spaces\\ \\ are\\ bad"
    )
    assert (
        utils.cygpath("\\\\server/share/spaces  are bad")
        == "//server/share/spaces  are bad"
    )
    assert (
        cyg_space("\\\\server/share/spaces  are bad")
        == "//server/share/spaces\\ \\ are\\ bad"
    )
    assert utils.cygpath("/C/test/spaces  are bad") == "/C/test/spaces  are bad"
    assert cyg_space("/C/test/spaces  are bad") == "/C/test/spaces\\ \\ are\\ bad"


def test_natrual_sort():
    items = ["test10", "test1", "Test3", "test2"]
    # Double check that our test doesn't sort naturally by default
    assert sorted(items) == ["Test3", "test1", "test10", "test2"]

    # Test that natural sort ignores case and groups numbers correctly
    result = utils.natural_sort(items)
    assert result == ["test1", "test2", "Test3", "test10"]

    # Test natural sorting using a custom sort key
    class Node:
        def __init__(self, name):
            super().__init__()
            self.name = name

    nodes = [Node(item) for item in items]
    result = utils.natural_sort(nodes, key=lambda i: i.name)
    check = [n.name for n in result]
    assert check == ["test1", "test2", "Test3", "test10"]


def test_star_import():
    """Check if a import was removed from __init__.py but did not get removed
    from `__all__`.

    Flake8 doesn't seem to capture F822 when run on `__init__.py`. Manually
    attempt a `from hab import *` import to ensure that all of the items listed
    in `__all__` are actually importable.
    """

    # https://stackoverflow.com/a/43059528
    import importlib

    # get a handle on the module
    mdl = importlib.import_module("hab")
    # is there an __all__?  if so respect it
    names = mdl.__dict__["__all__"]

    # Simulate `from hab import *` which can only be done at the module level.
    for k in names:
        # NOTE: If an exception is raised here, you need to remove the attribute
        # name from __all__ or make sure to import the missing object.
        # `AttributeError: module 'hab' has no attribute 'DistroVersion'`
        getattr(mdl, k)


def test_clear_caches(resolver):
    """Test that Resolver.clear_cache works as expected."""
    # Resolver cache is empty
    assert resolver._configs is None
    assert resolver._distros is None

    # Populate resolver cache data
    resolver.resolve("not_set")
    assert isinstance(resolver._configs, dict)
    assert isinstance(resolver._distros, dict)
    assert len(resolver._configs) > 1
    assert len(resolver._distros) > 1

    # Calling clear_caches resets the resolver cache
    resolver.clear_caches()
    assert resolver._configs is None
    assert resolver._distros is None


def test_uri_validate(config_root):
    """Test the `hab.uri.validate` entry_point."""
    resolver = Resolver(
        site=Site(
            [
                config_root / "site" / "site_ep_uri_validate.json",
                config_root / "site_main.json",
            ]
        )
    )

    # Test if an entry_point raises an exception
    with pytest.raises(
        Exception, match=r'URI "raise-error" was used, raising an exception.'
    ):
        cfg = resolver.resolve("raise-error")

    # "project_a" should be lower cased by the first validator.
    cfg = resolver.resolve("pRoJect_A/CammelCase")
    assert cfg.uri == "project_a/CammelCase"

    # "project_b" should be upper cased by the second validator.
    cfg = resolver.resolve("pRoJect_B/CammelCase")
    assert cfg.uri == "PROJECT_B/CammelCase"

    # Other URI's are not modified by any of the validators.
    cfg = resolver.resolve("proJECT_c/CammelCase")
    assert cfg.uri == "proJECT_c/CammelCase"

    # Test that the default behavior does nothing.
    resolver = Resolver(site=Site([config_root / "site_main.json"]))
    assert resolver.resolve("pRoJect_A/CammelCase").uri == "pRoJect_A/CammelCase"
    assert resolver.resolve("pRoJect_B/CammelCase").uri == "pRoJect_B/CammelCase"
    assert resolver.resolve("proJECT_c/CammelCase").uri == "proJECT_c/CammelCase"
